#ifndef _MYMATH_H_DS_
#define _MYMATH_H_DS_

#include <math.h>

#define PRECISION   1e-5
#define EPS         1e-6        /* data type is float */
#define INFINITY    FLT_MAX

#define ABS(X)			((X) > 0 ? (X) : (-(X)))
#define EQ_EPS(X, Y)	(ABS((X) - (Y)) < EPS)
#define EQ_ZERO(x)		(ABS(x) < EPS)
#define MIN(X, Y)		((X) > (Y)) ? (Y) : (X)
#define MAX(X, Y)		((X) > (Y)) ? (X) : (Y)
#define SIGN(X)			((X) > 0)

typedef struct float3	{ float x, y, z   ; }		float3;
typedef struct float4	{ float x, y, z, w; }		float4;
typedef struct float4x4 { float4 m0, m1, m2, m3; }	float4x4;

inline float3 make_float3(float x, float y, float z)
{ float3 a; a.x = x; a.y = y; a.z = z; return a; }
inline float dot(float3 a, float3 b) { return a.x * b.x + a.y * b.y + a.z * b.z; }
inline float3 cross(float3 a, float3 b)
{ return make_float3(a.y*b.z - a.z*b.y, a.z*b.x - a.x*b.z, a.x*b.y - a.y*b.x); }
inline float4 make_float4(float x, float y, float z, float w)
{ float4 a; a.x = x; a.y = y; a.z = z; a.w = w; return a; }
inline float dot(float4 a, float4 b) { return a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w; }
inline float4 operator+(float4 a, float4 b)
{ return make_float4(a.x+b.x, a.y+b.y, a.z+b.z, a.w+b.w); }
inline void operator+=(float4 &a, float4 b)
{ a.x+=b.x; a.y+=b.y; a.z+=b.z; a.w+=b.w; }
inline float4 operator-(float4 a, float4 b)
{ return make_float4(a.x-b.x, a.y-b.y, a.z-b.z, a.w-b.w); }
inline void operator-=(float4 &a, float4 b)
{ a.x-=b.x; a.y-=b.y; a.z-=b.z; a.w-=b.w; }
inline float4 operator*(float4 a, float4 b)
{ return make_float4(a.x*b.x, a.y*b.y, a.z*b.z, a.w*b.w); }
inline float4 operator*(float4 a, float b)
{ return make_float4(a.x*b, a.y*b, a.z*b, a.w*b); }
inline float4 operator*(float a, float4 b)
{ return make_float4(a*b.x, a*b.y, a*b.z, a*b.w); }
inline void operator*=(float4 &a, float b)
{ a.x*=b; a.y*=b; a.z*=b; a.w*=b; }
inline float4 _cross(float4 a, float4 b)
{ return make_float4(a.y*b.z - a.z*b.y, a.z*b.x - a.x*b.z, a.x*b.y - a.y*b.x, 1.f); }
inline float length(float3 v)
{
	return sqrtf(v.x*v.x + v.y*v.y + v.z*v.z);
}
inline float3 normalize(float3 v) 
{
	float _dot = dot(v, v);
//	if (EQ_ZERO(_dot)) { printf("Err!"); return v; } 
	if (EQ_ZERO(_dot)) { return v; } 
	float rsqrtf = 1 / sqrtf(_dot);
	return make_float3(v.x*rsqrtf, v.y*rsqrtf, v.z *rsqrtf);
}
inline void transpose(float4x4 &a)
{
	float4x4 b;
	b.m0 = make_float4(a.m0.x, a.m1.x, a.m2.x, a.m3.x);
	b.m1 = make_float4(a.m0.y, a.m1.y, a.m2.y, a.m3.y);
	b.m2 = make_float4(a.m0.z, a.m1.z, a.m2.z, a.m3.z);
	b.m3 = make_float4(a.m0.w, a.m1.w, a.m2.w, a.m3.w);
	a = b;
}
inline float4 mul_(float4x4 mat_, float4 vec)
{ return make_float4(dot(mat_.m0,vec),dot(mat_.m1,vec),dot(mat_.m2,vec),dot(mat_.m3,vec)); }
inline float4 mul(float4x4 mat, float4 vec)
{ transpose(mat); return mul_(mat, vec); }
inline float4x4 mul(float4x4 a, float4x4 b)
{
	float4x4 c; transpose(a);
	c.m0 = mul_(a, b.m0); c.m1 = mul_(a, b.m1); c.m2 = mul_(a, b.m2); c.m3 = mul_(a, b.m3);
	return c;
}

typedef struct Quaternion { float x, y, z, w; } Quaternion;
inline Quaternion float4toQuat(float4 a)
{ Quaternion b; b.x = a.x; b.y = a.y; b.z = a.z; b.w = a.w; return b; }
inline Quaternion add(Quaternion q1, Quaternion q2)
{
	Quaternion res;
	res.w = q1.w + q2.w; res.x = q1.x + q2.x; res.y = q1.y + q2.y; res.z = q1.z + q2.z;
	return res;
}
inline Quaternion mul(Quaternion q1, Quaternion q2)
{
	Quaternion res;
	res.w = q1.w*q2.w - (q1.x*q2.x + q1.y*q2.y + q1.z*q2.z);
	res.x = q1.w*q2.x + q2.w*q1.x + q1.y*q2.z - q2.y*q1.z;
	res.y = q1.w*q2.y + q2.w*q1.y + q2.x*q1.z - q1.x*q2.z;
	res.z = q1.w*q2.z + q2.w*q1.z + q1.x*q2.y - q2.x*q1.y;
	return res;
}
inline Quaternion convertQuat(float angle, float axisX, float axisY, float axisZ)
{
	Quaternion q;
	float3 axis = make_float3(axisX, axisY, axisZ);	axis = normalize(axis);
	float _sin = sin(angle*0.5);
	axis.x *= _sin; axis.y *= _sin; axis.z *= _sin; 
	q.x = axis.x; q.y = axis.y;	q.z = axis.z; q.w = cos(angle*0.5);
	return q;
}
inline void QuatToRotM(Quaternion q, float4x4 &rotMat)
{
	float4x4 r1, r2;
	r1.m0 = make_float4(q.x,  q.y,  q.z,  q.w);		r2.m0 = make_float4(q.x, q.w,  q.z,  -q.y);
	r1.m1 = make_float4(q.w,  q.z,  -q.y, -q.x);	r2.m1 = make_float4(q.y, -q.z, q.w,  q.x);
	r1.m2 = make_float4(-q.z, q.w,  q.x,  -q.y);	r2.m2 = make_float4(q.z, q.y,  -q.x, q.w);
	r1.m3 = make_float4(q.y,  -q.x, q.w,  -q.z);	r2.m3 = make_float4(q.w, -q.x, -q.y, -q.z);
	rotMat = mul(r1, r2);
}

#endif // _MYMATH_H_DS_